"""
Copyright © 2014-2022 Stb-tester.com Ltd.
Copyright © 2014 YouView TV Ltd.

License: LGPL v2.1 or (at your option) any later version (see
https://github.com/stb-tester/stb-tester/blob/master/LICENSE for details).
"""

from __future__ import annotations

import warnings
from typing import Optional

import cv2

from .config import get_config
from .imgutils import (
    crop, Frame, _frame_repr, _image_region, pixel_bounding_box)
from .logging import debug, ImageLogger
from .mask import load_mask, MaskTypes
from .types import Region


def is_screen_black(frame: Optional[Frame] = None,
                    mask: MaskTypes = Region.ALL,
                    threshold: Optional[int] = None,
                    region: Region = Region.ALL) -> _IsScreenBlackResult:
    """Check for the presence of a black screen in a video frame.

    :param Frame frame:
      If this is specified it is used as the video frame to check; otherwise a
      new frame is grabbed from the device-under-test. This is an image in
      OpenCV format (for example as returned by `frames` and `get_frame`).

    :param str|numpy.ndarray|Mask|Region mask:
        A `Region` or a mask that specifies which parts of the image to
        analyse. This accepts anything that can be converted to a Mask using
        `stbt.load_mask`. See :doc:`masks`.

    :param int threshold:
      Even when a video frame appears to be black, the intensity of its pixels
      is not always 0. To differentiate almost-black from non-black pixels, a
      binary threshold is applied to the frame. The ``threshold`` value is in
      the range 0 (black) to 255 (white). The global default (20) can be
      changed by setting ``threshold`` in the ``[is_screen_black]`` section of
      :ref:`.stbt.conf`.

    :param Region region:
      Deprecated synonym for ``mask``. Use ``mask`` instead.

    :returns:
        An object that will evaluate to true if the frame was black, or false
        if not black. The object has the following attributes:

        * **black** (*bool*) – True if the frame was black.
        * **frame** (`stbt.Frame`) – The video frame that was analysed.

    Changed in v33: ``mask`` accepts anything that can be converted to a Mask
    using `load_mask`. The ``region`` parameter is deprecated; pass your
    `Region` to ``mask`` instead. You can't specify ``mask`` and ``region``
    at the same time.
    """
    if threshold is None:
        threshold = get_config('is_screen_black', 'threshold', type_=int)

    if frame is None:
        from stbt_core import get_frame
        frame = get_frame()

    if region is not Region.ALL:
        if mask is not Region.ALL:
            raise ValueError("Cannot specify mask and region at the same time")
        warnings.warn(
            "stbt.is_screen_black: The 'region' parameter is deprecated; "
            "pass your Region to 'mask' instead",
            DeprecationWarning, stacklevel=2)
        mask = region

    mask_, region = load_mask(mask).to_array(_image_region(frame))

    imglog = ImageLogger("is_screen_black", region=region, threshold=threshold)
    imglog.imwrite("source", frame)

    grayframe = cv2.cvtColor(crop(frame, region), cv2.COLOR_BGR2GRAY)
    if mask_ is not None:
        imglog.imwrite("mask", mask_)
        cv2.bitwise_and(grayframe, mask_, dst=grayframe)
    maxVal = grayframe.max()

    result = _IsScreenBlackResult(bool(maxVal <= threshold), frame)
    debug("is_screen_black: {found} black screen using mask={mask}, "
          "threshold={threshold}: {result}, maximum_intensity={maxVal}".format(
              found="Found" if result.black else "Didn't find",
              mask=mask,
              threshold=threshold,
              result=result,
              maxVal=maxVal))

    if imglog.enabled:
        imglog.imwrite("gray", grayframe)
        _, thresholded = cv2.threshold(grayframe, threshold, 255,
                                       cv2.THRESH_BINARY)
        imglog.imwrite("non_black", thresholded)
        if result.black:
            bounding_box = None
        else:
            bounding_box = pixel_bounding_box(thresholded).translate(region)
        imglog.set(maxVal=maxVal, non_black_region=bounding_box)
    _log_image_debug(imglog, result)

    return result


class _IsScreenBlackResult():
    def __init__(self, black: bool, frame: Frame):
        self.black: bool = black
        self.frame: Frame = frame

    def __bool__(self):
        return self.black

    def __repr__(self):
        return ("_IsScreenBlackResult(black=%r, frame=%s)" % (
            self.black,
            _frame_repr(self.frame)))


def _log_image_debug(imglog, result):
    if not imglog.enabled:
        return

    template = """\
        <h4>is_screen_black: {{result.black}}</h4>

        {{ annotated_image(non_black_region) }}

        {% if "mask" in images %}
        <h5>Mask:</h5>
        <img src="mask.png" />
        {% endif %}

        <h5>Grayscale, masked:</h5>
        <img src="gray.png">
        <ul>
          <li>Maximum pixel intensity: {{maxVal}}
          <li>threshold={{threshold}}
        </ul>

        {% if not result.black %}
        <h5>Non-black pixels in region:</h5>
        <img src="non_black.png" />
        {% endif %}
    """

    imglog.html(template, result=result)



"""Detection & frame-accurate measurement of animations and transitions.

For example a selection that moves from one menu item to another or loading a
new screen such as a Guide and waiting for it to populate fully.

Because we want these measurements to be frame-accurate, we don't do expensive
image processing, relying instead on diffs between successive frames.

Copyright 2017-2018 Stb-tester.com Ltd.
License: LGPL v2.1 or (at your option) any later version (see
https://github.com/stb-tester/stb-tester/blob/master/LICENSE for details).
"""

from __future__ import annotations

import enum
import warnings
from typing import Iterator, Optional

from .diff import BGRDiff, Differ
from .imgutils import FrameT
from .logging import ddebug, debug, draw_on, warn
from .mask import MaskTypes
from .motion import DetectMotion
from .types import KeyT, Region, SizeT


def press_and_wait(
    key: KeyT,
    mask: MaskTypes = Region.ALL,
    region: Region = Region.ALL,
    timeout_secs: float = 10,
    stable_secs: float = 1,
    min_size: Optional[SizeT] = None,
    retries: int = 0,
    frames: Optional[Iterator[FrameT]] = None,
    _dut=None,
) -> Transition:

    """Press a key, then wait for the screen to change, then wait for it to stop
    changing.

    This can be used to wait for a menu selection to finish moving before
    attempting to OCR at the selection's new position; or to measure the
    duration of animations; or to measure how long it takes for a screen (such
    as an EPG) to finish populating.

    :param key: The name of the key to press (passed to `stbt.press`).

    :param mask:
        A `Region` or a mask that specifies which parts of the image to
        analyse. This accepts anything that can be converted to a Mask using
        `stbt.load_mask`. See :doc:`masks`.

    :param region:
        Deprecated synonym for ``mask``. Use ``mask`` instead.

    :param timeout_secs: A timeout in seconds. This function will return a
        falsey value if the transition didn't complete within this number of
        seconds from the key-press.

    :param stable_secs: A duration in seconds. The screen must stay unchanged
        (within the specified region or mask) for this long, for the transition
        to be considered "complete".

    :param min_size: A tuple of ``(width, height)``, in pixels, for differences
        to be considered as "motion". Use this to ignore small differences,
        such as the blinking text cursor in a search box.

    :param retries: Press the key again (up to this number of times) if the
        first press didn't have any effect (that is, if the status would have
        been `TransitionStatus.START_TIMEOUT`). Defaults to 0 (no retries).

    :param frames: An iterable of video-frames to analyse. Defaults to
        ``stbt.frames()``.

    :returns:
        A `Transition` object that will evaluate to true if the transition
        completed, false otherwise.

    Changed in v32: Use the same difference-detection algorithm as
    `wait_for_motion`.

    Added in v33: The ``started``, ``complete`` and ``stable`` attributes of
    the returned value.

    Changed in v33: ``mask`` accepts anything that can be converted to a Mask
    using `load_mask`. The ``region`` parameter is deprecated; pass your
    `Region` to ``mask`` instead. You can't specify ``mask`` and ``region``
    at the same time.

    Added in v34: The ``retries`` and ``frames`` parameters.

    Changed in v34: The difference-detection algorithm takes color into
    account.
    """
    if _dut is None:
        import stbt_core
        _dut = stbt_core
    if frames is None:
        frames = _dut.frames()

    if region is not Region.ALL:
        if mask is not Region.ALL:
            raise ValueError("Cannot specify mask and region at the same time")
        warnings.warn(
            "stbt.press_and_wait: The 'region' parameter is deprecated; "
            "pass your Region to 'mask' instead",
            DeprecationWarning, stacklevel=2)
        mask = region

    result = _press_and_wait(key, mask, timeout_secs, stable_secs,
                             min_size, frames, _dut)
    for i in range(retries):
        if result.status != TransitionStatus.START_TIMEOUT:
            return result
        warn("Keypress %s had no effect; retrying %i/%i", key, i + 1, retries)
        result = _press_and_wait(key, mask, timeout_secs, stable_secs,
                                 min_size, frames, _dut)
    return result


def _press_and_wait(key, mask, timeout_secs, stable_secs, min_size,
                    frames, _dut) -> Transition:

    t = _Transition(mask, timeout_secs, stable_secs, min_size, frames)
    press_result = _dut.press(key)
    debug("transition: %.3f: Pressed %s" % (press_result.end_time, key))
    result = t.wait(press_result)
    debug("press_and_wait(%r) -> %s" % (key, result))
    return result


press_and_wait.differ: Differ = BGRDiff()


def wait_for_transition_to_end(
    initial_frame: Optional[FrameT] = None,
    mask: MaskTypes = Region.ALL,
    region: Region = Region.ALL,
    timeout_secs: float = 10,
    stable_secs: float = 1,
    min_size: Optional[SizeT] = None,
    frames: Optional[Iterator[FrameT]] = None,
) -> Transition:

    """Wait for the screen to stop changing.

    In most cases you should use `press_and_wait` to measure a complete
    transition, but if you need to measure several points during a single
    transition you can use `wait_for_transition_to_end` as the last
    measurement. For example::

        keypress = stbt.press("KEY_OK")  # Launch my app
        m = stbt.wait_for_match("my-app-home-screen.png")
        time_to_first_frame = m.time - keypress.start_time
        end = wait_for_transition_to_end(m.frame)
        time_to_fully_populated = end.end_time - keypress.start_time

    :param initial_frame: The frame of video when the transition started. If
        `None`, we'll pull a new frame from the device under test.

    :param mask: See `press_and_wait`.
    :param region: See `press_and_wait`.
    :param timeout_secs: See `press_and_wait`.
    :param stable_secs: See `press_and_wait`.
    :param frames: See `press_and_wait`.

    :returns: See `press_and_wait`.
    """
    if frames is None:
        import stbt_core
        frames = stbt_core.frames()

    if region is not Region.ALL:
        if mask is not Region.ALL:
            raise ValueError("Cannot specify mask and region at the same time")
        warnings.warn(
            "stbt.wait_for_transition_to_end: The 'region' parameter is "
            "deprecated; pass your Region to 'mask' instead",
            DeprecationWarning, stacklevel=2)
        mask = region

    t = _Transition(mask, timeout_secs, stable_secs, min_size, frames)
    result = t.wait_for_transition_to_end(initial_frame)
    debug("wait_for_transition_to_end() -> %s" % (result,))
    return result


class _Transition():
    def __init__(self, mask: MaskTypes, timeout_secs: float,
                 stable_secs: float, min_size: SizeT | None,
                 frames: Iterator[FrameT]):
        self.mask = mask
        self.timeout_secs = timeout_secs
        self.stable_secs = stable_secs
        self.min_size = min_size
        self.frames = frames

        self.expiry_time = None

    def wait(self, press_result):
        self.expiry_time = press_result.end_time + self.timeout_secs

        differ = press_and_wait.differ.replace(min_size=self.min_size)
        dm = DetectMotion(differ, press_result.frame_before, self.mask)

        # Wait for animation to start
        for f in self.frames:
            if f.time < press_result.end_time:
                # Discard frame to work around latency in video-capture pipeline
                continue
            motion_result = dm.diff(f)
            draw_on(f, motion_result, label="transition")
            if motion_result:
                _debug("Animation started", f)
                animation_start_time = f.time
                break
            _debug("No change", f)
            if f.time >= self.expiry_time:
                _debug(
                    "Transition didn't start within %s seconds of pressing %s",
                    f, self.timeout_secs, press_result.key)
                return Transition(
                    press_result.key, f, TransitionStatus.START_TIMEOUT,
                    press_result.end_time, None, None)

        end_result = self.wait_for_transition_to_end(f)  # pylint:disable=undefined-loop-variable
        return Transition(
            press_result.key, end_result.frame, end_result.status,
            press_result.end_time, animation_start_time, end_result.end_time)

    def wait_for_transition_to_end(self, initial_frame):
        if initial_frame is None:
            initial_frame = next(self.frames)
        if self.expiry_time is None:
            self.expiry_time = initial_frame.time + self.timeout_secs

        first_stable_frame = initial_frame
        differ = press_and_wait.differ.replace(min_size=self.min_size)
        dm = DetectMotion(differ, initial_frame, self.mask)
        while True:
            f = next(self.frames)
            motion_result = dm.diff(f)
            draw_on(f, motion_result, label="transition")
            if motion_result:
                _debug("Animation in progress", f)
                first_stable_frame = f
            else:
                _debug("No change since previous frame", f)
            if f.time - first_stable_frame.time >= self.stable_secs:
                _debug("Transition complete (stable for %ss since %.3f).",
                       first_stable_frame, self.stable_secs,
                       first_stable_frame.time)
                return Transition(
                    None, first_stable_frame, TransitionStatus.COMPLETE,
                    None, initial_frame.time, first_stable_frame.time)
            if f.time >= self.expiry_time:
                _debug("Transition didn't end within %s seconds",
                       f, self.timeout_secs)
                return Transition(
                    None, f, TransitionStatus.STABLE_TIMEOUT,
                    None, initial_frame.time, None)


def _debug(s, f, *args):
    debug(("transition: %.3f: " + s) % ((getattr(f, "time", 0),) + args))


def _ddebug(s, f, *args):
    ddebug(("transition: %.3f: " + s) % ((getattr(f, "time", 0),) + args))


class Transition():
    """The return value from `press_and_wait` and `wait_for_transition_to_end`.

    This object will evaluate to true if the transition completed, false
    otherwise. It has the following attributes:

    :ivar str key: The name of the key that was pressed.
    :ivar stbt.Frame frame: If successful, the first video frame when
      the transition completed; if timed out, the last frame seen.
    :ivar stbt.TransitionStatus status: Either ``START_TIMEOUT`` (the
      transition didn't start – nothing moved), ``STABLE_TIMEOUT`` (the
      transition didn't end – movement didn't stop), or ``COMPLETE`` (the
      transition started and then stopped). If it's ``COMPLETE``, the whole
      object will evaluate as true.
    :ivar bool started: The transition started (movement was seen
      after the keypress). Implies that ``status`` is either ``COMPLETE``
      or ``STABLE_TIMEOUT``.
    :ivar bool complete: The transition completed (movement started
      and then stopped). Implies that ``status`` is ``COMPLETE``.
    :ivar bool stable: The screen is stable (no movement). Implies
      ``complete or not started``.
    :ivar float press_time: When the key-press completed.
    :ivar float animation_start_time: When animation started after the
      key-press (or ``None`` if timed out).
    :ivar float end_time: When animation completed (or ``None`` if timed out).
    :ivar float duration: Time from ``press_time`` to ``end_time`` (or
      ``None`` if timed out).
    :ivar float animation_duration: Time from ``animation_start_time``
      to ``end_time`` (or ``None`` if timed out).

    All times are measured in seconds since 1970-01-01T00:00Z; the
    timestamps can be compared with system time (the output of
    `time.time`).
    """
    def __init__(self, key, frame, status, press_time, animation_start_time,
                 end_time):
        self.key: KeyT = key
        self.frame: FrameT = frame
        self.status: TransitionStatus = status
        self.press_time: float = press_time
        self.animation_start_time: float = animation_start_time
        self.end_time: float | None = end_time

    def __repr__(self):
        return (
            "_TransitionResult(key=%r, frame=<Frame>, status=%s, "
            "press_time=%s, animation_start_time=%s, end_time=%s)" % (
                self.key,
                self.status,
                self.press_time,
                self.animation_start_time,
                self.end_time))

    def __str__(self):
        # Also lists the properties -- it's useful to see them in the logs.
        return (
            "_TransitionResult(key=%r, frame=<Frame>, status=%s, "
            "press_time=%s, animation_start_time=%s, end_time=%s, duration=%s, "
            "animation_duration=%s)" % (
                self.key,
                self.status,
                self.press_time,
                self.animation_start_time,
                self.end_time,
                self.duration,
                self.animation_duration))

    def __bool__(self):
        return self.status == TransitionStatus.COMPLETE

    @property
    def duration(self) -> Optional[float]:
        if self.end_time is None or self.press_time is None:
            return None
        return self.end_time - self.press_time

    @property
    def animation_duration(self) -> Optional[float]:
        if self.end_time is None or self.animation_start_time is None:
            return None
        return self.end_time - self.animation_start_time

    @property
    def started(self) -> bool:
        return self.status != TransitionStatus.START_TIMEOUT

    @property
    def complete(self) -> bool:
        return self.status == TransitionStatus.COMPLETE

    @property
    def stable(self) -> bool:
        return self.status in (TransitionStatus.START_TIMEOUT,
                               TransitionStatus.COMPLETE)


# For backwards compatibility with users who may have done
# `from _stbt.transition import TransitionResult`; remove in v35.
_TransitionResult = Transition


class TransitionStatus(enum.Enum):

    #: The transition didn't start (nothing moved).
    START_TIMEOUT = 0

    #: The transition didn't end (movement didn't stop).
    STABLE_TIMEOUT = 1

    #: The transition started and then stopped.
    COMPLETE = 2



from __future__ import annotations

import warnings
from collections import deque
from typing import Iterator, Optional

from .config import ConfigurationError, get_config
from .diff import BGRDiff, Differ, MotionResult
from .imgutils import _image_region, limit_time, FrameT
from .logging import debug, draw_on
from .mask import MaskTypes
from .types import Region, UITestFailure


def detect_motion(
    timeout_secs: float = 10,
    noise_threshold: Optional[int] = None,
    mask: MaskTypes = Region.ALL,
    region: Region = Region.ALL,
    frames: Optional[Iterator[FrameT]] = None,
) -> Iterator[MotionResult]:
    """Generator that yields a sequence of one `MotionResult` for each frame
    processed from the device-under-test's video stream.

    The `MotionResult` indicates whether any motion was detected.

    Use it in a ``for`` loop like this::

        for motionresult in stbt.detect_motion():
            ...

    In most cases you should use `wait_for_motion` instead.

    :type timeout_secs: int or float or None
    :param timeout_secs:
        A timeout in seconds. After this timeout the iterator will be exhausted.
        Thas is, a ``for`` loop like ``for m in detect_motion(timeout_secs=10)``
        will terminate after 10 seconds. If ``timeout_secs`` is ``None`` then
        the iterator will yield frames forever. Note that you can stop
        iterating (for example with ``break``) at any time.

    :param int noise_threshold:
        The difference in pixel intensity to ignore. Valid values range from 0
        (any difference is considered motion) to 255 (which would never report
        motion).

        This defaults to 25. You can override the global default value by
        setting ``noise_threshold`` in the ``[motion]`` section of
        :ref:`.stbt.conf`.

    :param str|numpy.ndarray|Mask|Region mask:
        A `Region` or a mask that specifies which parts of the image to
        analyse. This accepts anything that can be converted to a Mask using
        `stbt.load_mask`. See :doc:`masks`.

    :param Region region:
      Deprecated synonym for ``mask``. Use ``mask`` instead.

    :param Iterator[stbt.Frame] frames: An iterable of video-frames to analyse.
        Defaults to ``stbt.frames()``.

    Changed in v33: ``mask`` accepts anything that can be converted to a Mask
    using `load_mask`. The ``region`` parameter is deprecated; pass your
    `Region` to ``mask`` instead. You can't specify ``mask`` and ``region``
    at the same time.

    Changed in v34: The difference-detection algorithm takes color into
    account. The ``noise_threshold`` parameter changed range (from 0.0-1.0 to
    0-255), sense (from "bigger is stricter" to "smaller is stricter"), and
    default value (from 0.84 to 25).
    """
    if frames is None:
        import stbt_core
        frames = stbt_core.frames()

    frames = limit_time(frames, timeout_secs)  # pylint: disable=redefined-variable-type

    if region is not Region.ALL:
        if mask is not Region.ALL:
            raise ValueError("Cannot specify mask and region at the same time")
        warnings.warn(
            "stbt.detect_motion: The 'region' parameter is deprecated; "
            "pass your Region to 'mask' instead",
            DeprecationWarning, stacklevel=2)
        mask = region

    if noise_threshold is None:
        noise_threshold = get_config('motion', 'noise_threshold', type_=int)

    debug(f"Searching for motion, using mask={mask}")

    try:
        frame = next(frames)
    except StopIteration:
        return

    differ = detect_motion.differ.replace(threshold=noise_threshold)
    dm = DetectMotion(differ, frame, mask)
    for frame in frames:
        result = dm.diff(frame)
        draw_on(frame, result, label="detect_motion()")
        debug("%s found: %s" % (
            "Motion" if result.motion else "No motion", str(result)))
        yield result


detect_motion.differ : Differ = BGRDiff()


class DetectMotion():
    """Concrete implementation of the `detect_motion` algorithm.

    This class is an internal implementation detail, not part of the public
    API. It isn't part of `detect_motion` because it offers a more convenient
    API for `press_and_wait` to use.

    Responsibilities of this class:

    - Remember the work done on already-seen frames (e.g. GrayscaleDiff's
      colorspace conversion).
    - The logic for when we update the "reference" frame.
    """
    def __init__(self, differ: Differ, initial_frame: FrameT,
                 mask: MaskTypes = Region.ALL):
        self.differ: Differ = differ
        self.mask_tuple = differ.preprocess_mask(
            mask, _image_region(initial_frame))
        self.prev_frame = differ.preprocess(initial_frame, self.mask_tuple)

    def diff(self, frame: FrameT) -> MotionResult:
        new_frame = self.differ.preprocess(frame, self.mask_tuple)
        motion = self.differ.diff(self.prev_frame, new_frame, self.mask_tuple)

        if motion:
            # Only update the comparison frame if it's different to the previous
            # one.  This makes `detect_motion` more sensitive to slow motion
            # because the differences between frames 1 and 2 might be small and
            # the differences between frames 2 and 3 might be small but we'd see
            # the difference by looking between 1 and 3.
            self.prev_frame = new_frame

        return motion


def wait_for_motion(
    timeout_secs: float = 10,
    consecutive_frames: "Optional[int | str]" = None,
    noise_threshold: Optional[int] = None,
    mask: MaskTypes = Region.ALL,
    region: Region = Region.ALL,
    frames: "Optional[Iterator[FrameT]]" = None,
) -> MotionResult:
    """Search for motion in the device-under-test's video stream.

    "Motion" is difference in pixel values between two frames.

    :type timeout_secs: int or float or None
    :param timeout_secs:
        A timeout in seconds. This function will raise `MotionTimeout` if no
        motion is detected within this time.

    :type consecutive_frames: int or str
    :param consecutive_frames:
        Considers the video stream to have motion if there were differences
        between the specified number of consecutive frames. This can be:

        * a positive integer value, or
        * a string in the form "x/y", where "x" is the number of frames with
          motion detected out of a sliding window of "y" frames.

        This defaults to "10/20". You can override the global default value by
        setting ``consecutive_frames`` in the ``[motion]`` section of
        :ref:`.stbt.conf`.

    :param int noise_threshold: See `detect_motion`.
    :param str|numpy.ndarray|Mask|Region mask: See `detect_motion`.
    :param Region region: See `detect_motion`.
    :param Iterator[stbt.Frame] frames: See `detect_motion`.

    :returns: `MotionResult` when motion is detected. The MotionResult's
        ``time`` and ``frame`` attributes correspond to the first frame in
        which motion was detected.
    :raises: `MotionTimeout` if no motion is detected after ``timeout_secs``
        seconds.

    Changed in v33: ``mask`` accepts anything that can be converted to a Mask
    using `load_mask`. The ``region`` parameter is deprecated; pass your
    `Region` to ``mask`` instead. You can't specify ``mask`` and ``region``
    at the same time.

    Changed in v34: The difference-detection algorithm takes color into
    account. The ``noise_threshold`` parameter changed range (from 0.0-1.0 to
    0-255), sense (from "bigger is stricter" to "smaller is stricter"), and
    default value (from 0.84 to 25).
    """
    if frames is None:
        import stbt_core
        frames = stbt_core.frames()

    if consecutive_frames is None:
        consecutive_frames = get_config('motion', 'consecutive_frames')

    consecutive_frames = str(consecutive_frames)
    if '/' in consecutive_frames:
        motion_frames = int(consecutive_frames.split('/')[0])
        considered_frames = int(consecutive_frames.split('/')[1])
    else:
        motion_frames = int(consecutive_frames)
        considered_frames = int(consecutive_frames)

    if motion_frames > considered_frames:
        raise ConfigurationError(
            "`motion_frames` exceeds `considered_frames`")

    if region is not Region.ALL:
        if mask is not Region.ALL:
            raise ValueError("Cannot specify mask and region at the same time")
        warnings.warn(
            "stbt.wait_for_motion: The 'region' parameter is deprecated; "
            "pass your Region to 'mask' instead",
            DeprecationWarning, stacklevel=2)
        mask = region

    debug("Waiting for %d out of %d frames with motion, using mask=%r" % (
        motion_frames, considered_frames, mask))

    matches = deque(maxlen=considered_frames)
    motion_count = 0
    last_frame = None
    for res in detect_motion(
            timeout_secs, noise_threshold, mask, frames=frames):
        motion_count += bool(res)
        if len(matches) == matches.maxlen:
            motion_count -= bool(matches.popleft())
        matches.append(res)
        if motion_count >= motion_frames:
            debug("Motion detected.")
            # We want to return the first True motion result as this is when
            # the motion actually started.
            for result in matches:
                if result:
                    return result
            assert False, ("Logic error in wait_for_motion: This code "
                           "should never be reached")
        last_frame = res.frame

    raise MotionTimeout(last_frame, mask, timeout_secs)


class MotionTimeout(UITestFailure):
    """Exception raised by `wait_for_motion`.

    :ivar Frame screenshot: The last video frame that `wait_for_motion` checked
        before timing out.

    :vartype mask: Mask or None
    :ivar mask: The mask that was used, if any.

    :vartype timeout_secs: int or float
    :ivar timeout_secs: Number of seconds that motion was searched for.
    """
    def __init__(self, screenshot: FrameT, mask: MaskTypes,
                 timeout_secs: float):
        super().__init__()
        self.screenshot: FrameT = screenshot
        self.mask: MaskTypes = mask
        self.timeout_secs: float = timeout_secs

    def __str__(self):
        return "Didn't find motion within %g seconds, using mask=%r" % (
            self.timeout_secs, self.mask)



from __future__ import annotations

import typing

import cv2
import numpy

from .imgutils import Frame, FrameT, crop, _frame_repr, pixel_bounding_box
from .logging import debug, ddebug, ImageLogger
from .mask import load_mask, MaskTypes
from .types import Region, SizeT


class MotionResult():
    """The result from `detect_motion` and `wait_for_motion`.

    :ivar float time: The time at which the video-frame was captured, in
        seconds since 1970-01-01T00:00Z. This timestamp can be compared with
        system time (``time.time()``).

    :ivar bool motion: True if motion was found. This is the same as evaluating
        ``MotionResult`` as a bool. That is, ``if result:`` will behave the
        same as ``if result.motion:``.

    :ivar Region region: Bounding box where the motion was found, or ``None``
        if no motion was found.

    :ivar Frame frame: The video frame in which motion was (or wasn't) found.
    """
    _fields = ("time", "motion", "region", "frame")

    def __init__(self, time: float, motion: bool, region: Region, frame: Frame):
        self.time: float = time
        self.motion: bool = motion
        self.region: Region = region
        self.frame: Frame = frame

    def __bool__(self) -> bool:
        return self.motion

    def __nonzero__(self) -> bool:
        return self.__bool__()

    def __repr__(self) -> str:
        return (
            "MotionResult(time=%s, motion=%r, region=%r, frame=%s)" % (
                "None" if self.time is None else "%.3f" % self.time,
                self.motion, self.region, _frame_repr(self.frame)))


class UNSET:
    """Sentinel"""


class Differ:
    """An algorithm that compares two images or frames to find the differences
    between them.

    Subclasses of this class implement the actual diffing algorithms: See
    `BGRDiff` and `GrayscaleDiff`.
    """

    # This is a low level interface that allows us to customise how two images
    # or frames are compared. It's used by `detect_motion`/`wait_for_motion`
    # and `press_and_wait`.
    #
    # `Differ` is designed so we can efficiently implement `GrayscaleDiff`
    # without doing the colour -> grayscale conversion on the same image twice.
    # To avoid internal state the interface makes preprocessing explicit; the
    # caller (`DetectMotion`) is responsible for keeping track of that state.
    # This makes the classes hard to use, but that's ok — it's a low-level
    # interface intended mostly for internal use.
    #
    # The only public interface is the subclass constructors. The methods are
    # not intended for use by external code.

    _PreProcessedFrame: typing.TypeAlias
    _PreProcessedMask: typing.TypeAlias

    def replace(self, min_size=UNSET, threshold=UNSET, erode=UNSET):
        """
        Return a new Differ with the specified parameters replaced.

        Differs are immutable. This method allows creating a new differ,
        overriding some of the parameters.

        This is needed to allow passing parameters like `min_size` when calling
        functions like `press_and_wait`.

        :meta private:
        """
        raise NotImplementedError(
            "%s.replace is not implemented" % self.__class__.__name__)

    def preprocess_mask(
            self, mask: MaskTypes,
            frame_region: Region  # pylint:disable=unused-argument
    ) -> "_PreProcessedMask":
        """
        Pre-process a mask.  The returned value from this will be passed to
        `preprocess` and `diff`.

        :meta private:
        """
        return mask

    def preprocess(
            self, frame: FrameT,
            mask: "_PreProcessedMask"  # pylint:disable=unused-argument
    ) -> "_PreProcessedFrame":
        """
        Pre-process a frame.  The returned value from this will be passed to
        `diff`.  `mask_tuple` is the return value from `preprocess_mask`.

        :meta private:
        """
        return frame

    def diff(self, a: "_PreProcessedFrame", b: "_PreProcessedFrame",
             mask: "_PreProcessedMask") -> MotionResult:
        """
        Compare two frames.  `a` and `b` are the return values from
        `preprocess`.  `mask_tuple` is the return value from `preprocess_mask`.

        :meta private:
        """
        raise NotImplementedError(
            "%s.diff is not implemented" % self.__class__.__name__)


class BGRDiff(Differ):
    """Compares 2 frames by calculating the color distance between them.

    The algorithm calculates the euclidean distance in BGR colorspace between
    each pair of corresponding pixels in the 2 frames. This distance is then
    binarized using the specified threshold: Values smaller than the threshold
    are ignored. Then, an "erode" operation removes differences that are only 1
    pixel wide or high. If any differences remain, the 2 frames are considered
    different.

    This is the default diffing algorithm for `detect_motion`,
    `wait_for_motion`, `press_and_wait`, `find_selection_from_background`,
    and `ocr`'s `text_color`.
    """

    def __init__(
        self,
        min_size: SizeT | None = None,
        threshold: float = 25,
        erode: bool = True,
    ):
        self.min_size = min_size
        self.threshold = threshold

        if isinstance(erode, numpy.ndarray):  # For power users
            kernel = erode
        elif erode:
            kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
        else:
            kernel = None
        self.kernel = kernel

    def replace(self, min_size=UNSET, threshold=UNSET, erode=UNSET):
        if min_size is UNSET:
            min_size = self.min_size
        if threshold is UNSET:
            threshold = self.threshold
        if erode is UNSET:
            erode = self.kernel
        return self.__class__(min_size, threshold, erode)

    def preprocess_mask(
            self, mask: MaskTypes, frame_region: Region):
        return load_mask(mask).to_array(frame_region)

    def diff(self, a, b, mask):
        mask_pixels, region = mask
        prev_frame = a
        frame = b

        imglog = ImageLogger("BGRDiff", region=region,
                             min_size=self.min_size, threshold=self.threshold)
        imglog.imwrite("source", frame)
        imglog.imwrite("previous_frame", prev_frame)

        cframe = crop(frame, region)
        cprev = crop(prev_frame, region)

        d = _threshold_diff_bgr(cprev, cframe, (self.threshold ** 2) * 3,
                                imglog, mask_pixels)
        if mask_pixels is not None:
            numpy.bitwise_and(d, mask_pixels[:, :, 0], out=d)
            imglog.imwrite("mask", mask_pixels)

        if imglog.enabled:
            imglog.imwrite("thresholded", d * 255)

        if self.kernel is not None:
            d = cv2.morphologyEx(d, cv2.MORPH_OPEN, self.kernel)
            if imglog.enabled:
                imglog.imwrite("eroded", d * 255)

        out_region = pixel_bounding_box(d)
        if out_region:
            # Undo crop:
            out_region = out_region.translate(region)

        motion = bool(out_region and (
            self.min_size is None or
            (out_region.width >= self.min_size[0] and
             out_region.height >= self.min_size[1])))
        result = MotionResult(getattr(frame, "time", None), motion,
                              out_region, frame)
        ddebug(str(result))
        imglog.html(BGRDIFF_HTML, result=result)
        return result


def _threshold_diff_bgr(
        a: numpy.ndarray[numpy.uint8], b: numpy.ndarray[numpy.uint8],
        threshold: int,
        imglog: ImageLogger,
        mask_pixels: "numpy.ndarray[numpy.uint8] | None"
) -> numpy.ndarray[numpy.uint8]:

    if a.shape[:2] != b.shape[:2]:
        raise ValueError("Images must be the same size")
    if (len(a.shape) < 3 or a.shape[2] != 3 or
            len(b.shape) < 3 or b.shape[2] != 3):
        raise ValueError("Images must be 3-channel BGR images")
    try:
        from . import libstbt
        if not imglog.enabled:
            return libstbt.threshold_diff_bgr(a, b, threshold)
    except (ImportError, NotImplementedError) as e:
        debug("BGRDiff missed fast-path: %s" % e)

    return _threshold_diff_bgr_numpy(a, b, threshold, imglog, mask_pixels)


def _threshold_diff_bgr_numpy(
        a: numpy.ndarray[numpy.uint8], b: numpy.ndarray[numpy.uint8],
        threshold: int,
        imglog: "ImageLogger | None" = None,
        mask_pixels: "numpy.ndarray[numpy.uint8] | None" = None
) -> numpy.ndarray[numpy.uint8]:

    sqd = numpy.subtract(a, b, dtype=numpy.int32)
    sqd = (sqd[:, :, 0] ** 2 +
           sqd[:, :, 1] ** 2 +
           sqd[:, :, 2] ** 2)

    if imglog is not None and imglog.enabled:
        normalised = numpy.sqrt(sqd / 3)
        if mask_pixels is None:
            imglog.imwrite("sqd", normalised)
        else:
            imglog.imwrite("sqd",
                           normalised.astype(numpy.uint8) &
                           mask_pixels[:, :, 0])

    return (sqd >= threshold).astype(numpy.uint8)


BGRDIFF_HTML = """\
    <h4>
      BGRDiff:
      {{ "Found" if result.motion else "Didn't find" }} differences
    </h4>

    {{ annotated_image(result) }}

    <h5>Result:</h5>
    <pre><code>{{ result | escape }}</code></pre>

    {% if result.region and not result.motion %}
    <p>Found motion <code>{{ result.region | escape }}</code> smaller than
    min_size <code>{{ min_size | escape }}</code>.</p>
    {% endif %}

    <h5>Previous frame:</h5>
    <img src="previous_frame.png" />

    <h5>Differences:</h5>
    <img src="sqd.png" />

    {% if "mask" in images %}
    <h5>Mask:</h5>
    <img src="mask.png" />
    {% endif %}

    <h5>Differences above threshold ({{threshold}}):</h5>
    <img src="thresholded.png" />

    {% if "eroded" in images %}
    <h5>Eroded:</h5>
    <img src="eroded.png" />
    {% endif %}
"""


class GrayscaleDiff(Differ):
    """Compares 2 frames by converting them to grayscale, calculating
    pixel-wise absolute differences, and ignoring differences below a
    threshold.

    This was the default diffing algorithm for `wait_for_motion` and
    `press_and_wait` before v34.
    """

    def __init__(
        self,
        min_size: SizeT | None = None,
        threshold: float = 0.84,
        erode: bool = True,
    ):
        self.min_size = min_size
        self.threshold = threshold

        if isinstance(erode, numpy.ndarray):  # For power users
            kernel = erode
        elif erode:
            kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
        else:
            kernel = None
        self.kernel = kernel

    def replace(self, min_size=UNSET, threshold=UNSET, erode=UNSET):
        if min_size is UNSET:
            min_size = self.min_size
        if threshold is UNSET:
            threshold = self.threshold
        if erode is UNSET:
            erode = self.kernel
        return self.__class__(min_size, threshold, erode)

    def preprocess_mask(
            self, mask: MaskTypes, frame_region: Region):
        return load_mask(mask).to_array(frame_region)

    def preprocess(self, frame, mask):
        _, region = mask
        return frame, cv2.cvtColor(crop(frame, region), cv2.COLOR_BGR2GRAY)

    def diff(self, a, b, mask) -> MotionResult:
        _, prev_frame_gray = a
        frame, frame_gray = b
        mask, region = mask

        imglog = ImageLogger("GrayscaleDiff", region=region,
                             min_size=self.min_size,
                             threshold=self.threshold)
        imglog.imwrite("source", frame)
        imglog.imwrite("gray", frame_gray)
        imglog.imwrite("previous_frame_gray", prev_frame_gray)

        absdiff = cv2.absdiff(prev_frame_gray, frame_gray)
        imglog.imwrite("absdiff", absdiff)

        if mask is not None:
            absdiff = cv2.bitwise_and(absdiff, mask)
            imglog.imwrite("mask", mask)
            imglog.imwrite("absdiff_masked", absdiff)

        _, thresholded = cv2.threshold(
            absdiff, int((1 - self.threshold) * 255), 255,
            cv2.THRESH_BINARY)
        imglog.imwrite("absdiff_threshold", thresholded)
        if self.kernel is not None:
            thresholded = cv2.morphologyEx(
                thresholded, cv2.MORPH_OPEN, self.kernel)
            imglog.imwrite("absdiff_threshold_erode", thresholded)

        out_region = pixel_bounding_box(thresholded)
        if out_region:
            # Undo crop:
            out_region = out_region.translate(region)

        motion = bool(out_region and (
            self.min_size is None or
            (out_region.width >= self.min_size[0] and
             out_region.height >= self.min_size[1])))

        result = MotionResult(getattr(frame, "time", None), motion,
                              out_region, frame)
        ddebug(str(result))
        imglog.html(GRAYSCALEDIFF_HTML, result=result)
        return result


GRAYSCALEDIFF_HTML = """\
    <h4>
      GrayscaleDiff:
      {{ "Found" if result.motion else "Didn't find" }} differences
    </h4>

    {{ annotated_image(result) }}

    <h5>Result:</h5>
    <pre><code>{{ result | escape }}</code></pre>

    {% if result.region and not result.motion %}
    <p>Found motion <code>{{ result.region | escape }}</code> smaller than
    min_size <code>{{ min_size | escape }}</code>.</p>
    {% endif %}

    <h5>ROI Gray:</h5>
    <img src="gray.png" />

    <h5>Previous frame ROI Gray:</h5>
    <img src="previous_frame_gray.png" />

    <h5>Absolute difference:</h5>
    <img src="absdiff.png" />

    {% if "mask" in images %}
    <h5>Mask:</h5>
    <img src="mask.png" />
    <h5>Absolute difference – masked:</h5>
    <img src="absdiff_masked.png" />
    {% endif %}

    <h5>Binarized (threshold={{threshold}}):</h5>
    <img src="absdiff_threshold.png" />

    {% if "eroded" in images %}
    <h5>Eroded:</h5>
    <img src="absdiff_threshold_erode.png" />
    {% endif %}
"""



import numpy

from .logging import debug


def sqdiff(template, frame) -> "tuple[int, int]":
    if template.shape[:2] != frame.shape[:2]:
        raise ValueError("Template and frame must be the same size")
    try:
        from . import libstbt
        return libstbt.sqdiff(template, frame)
    except (ImportError, NotImplementedError) as e:
        debug("sqdiff Missed fast-path: %s" % e)
        return _sqdiff_numpy(template, frame)


def _sqdiff_numpy(template, frame):
    template = template.astype(numpy.int64)
    frame = frame.astype(numpy.int64)
    if template.shape[2] == 4:
        # Masked
        x = ((template[:, :, :3] - frame) ** 2)[template[:, :, 3] == 255]
    else:
        x = (template - frame) ** 2
    return numpy.sum(x), x.size


def _random_template(size=(1280, 720)):
    tsize = (numpy.random.randint(1, size[0] + 1),
             numpy.random.randint(1, size[1] + 1))
    toff = (numpy.random.randint(size[0] - tsize[0] + 1),
            numpy.random.randint(size[1] - tsize[1] + 1))

    f = numpy.random.randint(0, 256, (size[1], size[0], 3),
                             dtype=numpy.uint8)
    t = numpy.random.randint(0, 256, (tsize[1], tsize[0], 3),
                             dtype=numpy.uint8)
    tt = numpy.random.randint(0, 256, (tsize[1], tsize[0], 4),
                              dtype=numpy.uint8)
    mask = tt[:, :, 3]
    mask[mask & 1 == 1] = 255
    mask[mask < 255] = 0

    f_cropped = f[toff[1]:toff[1] + tsize[1], toff[0]:toff[0] + tsize[0], :]
    return f_cropped, t, tt


def test_sqdiff():
    from .libstbt import sqdiff as sqdiff_c
    f = numpy.array(range(1280 * 720 * 3), dtype=numpy.uint8)
    f.shape = (720, 1280, 3)
    t = numpy.zeros((720, 1280, 3), dtype=numpy.uint8)
    tt = numpy.zeros((720, 1280, 4), dtype=numpy.uint8)
    t[:, :, :] = f
    tt[:, :, :3] = f
    tt[:, :, 3] = 255

    assert (0, 1280 * 720 * 3) == sqdiff_c(t, f)
    assert (0, 1280 * 720 * 3) == sqdiff_c(tt, f)
    assert (0, 1280 * 720 * 3) == sqdiff_c(tt[:, :, :3], f)

    f = numpy.ones((720, 1280, 3), dtype=numpy.uint8) * 255
    t = numpy.zeros((720, 1280, 3), dtype=numpy.uint8)
    tt = numpy.zeros((720, 1280, 4), dtype=numpy.uint8)

    assert (1280 * 720 * 255 * 255 * 3, 1280 * 720 * 3) == sqdiff_c(t, f)
    assert (1280 * 720 * 255 * 255 * 3, 1280 * 720 * 3) == sqdiff_c(t, f)
    assert (0, 0) == sqdiff_c(tt, f)

    tt[:, :, 3] = 255
    assert (1280 * 720 * 255 * 255 * 3, 1280 * 720 * 3) == sqdiff_c(t, f)
    assert (1280 * 720 * 255 * 255 * 3, 1280 * 720 * 3) == sqdiff_c(t, f)
    assert (1280 * 720 * 255 * 255 * 3, 1280 * 720 * 3) == sqdiff_c(tt, f)


def test_sqdiff_c_numpy_equivalence():
    from .libstbt import sqdiff as sqdiff_c
    for _ in range(100):
        frame_cropped, template, template_transparent = _random_template()

        for t in (template, template_transparent,
                  template_transparent[:, :, :3]):
            assert (_sqdiff_numpy(t, frame_cropped) ==
                    sqdiff_c(t, frame_cropped))


def _make_sqdiff_numba():
    # numba implementation included for the purposes of comparison.
    try:
        import numba
    except ImportError:
        return None

    def _sqdiff_numba(template, frame):
        if template.shape[2] == 3:
            return _sqdiff_numba_nomask(template, frame)
        else:
            return _sqdiff_numba_masked(template, frame)

    @numba.jit(nopython=True, nogil=True)
    def _sqdiff_numba_nomask(template, frame):
        cum = 0
        s = template.shape
        for i in range(s[0]):
            for j in range(s[1]):
                for k in range(s[2]):
                    cum += (template[i, j, k] - frame[i, j, k]) ** 2
        return cum, frame.size

    @numba.jit(nopython=True, nogil=True)
    def _sqdiff_numba_masked(template, frame):
        # mask is either 0 or 255
        cum = 0
        maskcount = 0
        s = template.shape
        for i in range(s[0]):
            for j in range(s[1]):
                if template[i, j, 3] == 255:
                    for k in range(s[2]):
                        cum += (template[i, j, k] - frame[i, j, k]) ** 2
                    maskcount += 1
        return cum, maskcount * 3

    return _sqdiff_numba


def _measure_performance():
    import timeit
    from .libstbt import sqdiff as sqdiff_c

    _sqdiff_numba = _make_sqdiff_numba()

    print("All times in ms                         numpy\tnumba")
    print("type    \tnumpy\tnumba\tC\tspeedup\tspeedup\tsize\talignment")
    for _ in range(100):
        frame_cropped, template, template_transparent = _random_template()

        for l, t in [("template ", template),
                     ("with mask", template_transparent),
                     ("unmasked ", template_transparent[:, :, :3])]:
            # pylint: disable=cell-var-from-loop

            np_time = min(timeit.repeat(
                lambda: _sqdiff_numpy(t, frame_cropped),
                repeat=3, number=10)) / 10
            c_time = min(timeit.repeat(
                lambda: sqdiff_c(t, frame_cropped),
                repeat=3, number=10)) / 10
            if _sqdiff_numba:
                numba_time = min(timeit.repeat(
                    lambda: _sqdiff_numba(t, frame_cropped),
                    repeat=3, number=10)) / 10
            else:
                numba_time = float('nan')
            print("%s\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%i x %i \t%s" % (
                l, np_time * 1000, numba_time * 1000, c_time * 1000,
                np_time / c_time, numba_time / c_time,
                frame_cropped.shape[1], frame_cropped.shape[0],
                frame_cropped.ctypes.data % 8))
